@@ -1,6 +1,6 @@
 /* cmds.c
  *
- * Copyright (c) 1996-2000 Mike Gleason, NCEMRSoft.
+ * Copyright (c) 1996-2001 Mike Gleason, NCEMRSoft.
  * All rights reserved.
  *
  */
@@ -17,17 +17,19 @@ FTPChdir(const FTPCIPtr cip, const char *const cdCwd)
 	if (strcmp(cip->magic, kLibraryMagic))
 		return (kErrBadMagic);
 
-	if ((cdCwd == NULL) || (cdCwd[0] == '\0')) {
+	if (cdCwd == NULL) {
 		result = kErrInvalidDirParam;
 		cip->errNo = kErrInvalidDirParam;
 	} else {
-		if (strcmp(cdCwd, "..") == 0)
+		if (cdCwd[0] == '\0')	/* But allow FTPChdir(cip, ".") to go through. */
+			result = 2;	
+		else if (strcmp(cdCwd, "..") == 0)
 			result = FTPCmd(cip, "CDUP");
 		else
 			result = FTPCmd(cip, "CWD %s", cdCwd);
 		if (result >= 0) {
 			if (result == 2) {
-				result = 0;
+				result = kNoErr;
 			} else {
 				result = kErrCWDFailed;
 				cip->errNo = kErrCWDFailed;
@@ -140,11 +142,9 @@ FTPRmdirRecursiveL2(const FTPCIPtr cip)
 		} else {
 			/* Assume it was a file -- remove it. */
 			result = FTPDelete(cip, file, kRecursiveNo, kGlobNo);
-			if (result != kNoErr) {
-				/* We couldn't remove the file.
-				 * Try continuing to remove the rest.
-				 */
-			}
+			/* Try continuing to remove the rest,
+			 * even if this failed.
+			 */
 		}
 	}
 	DisposeLineListContents(&fileList);
@@ -321,10 +321,14 @@ FTPChdirAndGetCWD(const FTPCIPtr cip, const char *const cdCwd, char *const newCw
 	if (strcmp(cip->magic, kLibraryMagic))
 		return (kErrBadMagic);
 
-	if ((newCwd == NULL) || (cdCwd == NULL) || (cdCwd[0] == '\0')) {
+	if ((newCwd == NULL) || (cdCwd == NULL)) {
 		result = kErrInvalidDirParam;
 		cip->errNo = kErrInvalidDirParam;
 	} else {
+		if (cdCwd[0] == '\0') {	/* But allow FTPChdir(cip, ".") to go through. */
+			result = FTPGetCWD(cip, newCwd, newCwdSize);
+			return (result);
+		}
 		rp = InitResponse();
 		if (rp == NULL) {
 			result = kErrMallocFailed;
@@ -368,6 +372,294 @@ FTPChdirAndGetCWD(const FTPCIPtr cip, const char *const cdCwd, char *const newCw
 
 
 
+int
+FTPChdir3(FTPCIPtr cip, const char *const cdCwd, char *const newCwd, const size_t newCwdSize, int flags)
+{
+	char *cp, *startcp;
+	int result;
+	int lastSubDir;
+	int mkd, pwd;
+	int chdir_err;
+	int did_chdir, did_mkdir;
+
+	if (cip == NULL)
+		return (kErrBadParameter);
+	if (strcmp(cip->magic, kLibraryMagic))
+		return (kErrBadMagic);
+
+	if (cdCwd == NULL) {
+		result = kErrInvalidDirParam;
+		cip->errNo = kErrInvalidDirParam;
+		return result;
+	}
+
+	/* Retain backwards compatibility with earlier library versions. */
+	if (flags == kChdirOnly)
+		flags = kChdirFullPath;
+
+	chdir_err = kNoErr;
+	if ((flags & kChdirFullPath) != 0) {
+		did_chdir = 0;
+		did_mkdir = 0;
+		if ((flags & kChdirAndGetCWD) != 0) {
+			result = FTPChdirAndGetCWD(cip, cdCwd, newCwd, newCwdSize);
+			if (result == kNoErr) {
+				did_chdir = 1;
+			} else {
+				chdir_err = result;
+			}
+		} else {
+			result = FTPChdir(cip, cdCwd);
+			if (result == kNoErr) {
+				did_chdir = 1;
+			} else {
+				chdir_err = result;
+			}
+		}
+		if ((did_chdir == 0) && ((flags & kChdirAndMkdir) != 0)) {
+			result = FTPMkdir(cip, cdCwd, kRecursiveYes);
+			if (result == kNoErr) {
+				did_mkdir = 1;
+			}
+		}
+		/* If we just created the directory, chdir to it now. */
+		if ((did_mkdir != 0) && ((flags & kChdirAndGetCWD) != 0)) {
+			result = FTPChdirAndGetCWD(cip, cdCwd, newCwd, newCwdSize);
+			if (result == kNoErr) {
+				did_chdir = 1;
+			} else {
+				chdir_err = result;
+			}
+		} else if (did_mkdir != 0) {
+			result = FTPChdir(cip, cdCwd);
+			if (result == kNoErr) {
+				did_chdir = 1;
+			} else {
+				chdir_err = result;
+			}
+		}
+		if (did_chdir != 0)
+			return (kNoErr);
+	}
+
+	if ((flags & kChdirOneSubdirAtATime) == 0) {
+		/* Valid flags must have one or both of 
+		 * kChdirOneSubdirAtATime and kChdirFullPath set.
+		 * If we're here, subdir mode was not requested,
+		 * and if chdir_err is 0 then kChdirFullPath was
+		 * not set either.
+		 */
+		if (chdir_err != kNoErr)
+			return (chdir_err);
+		return (kErrBadParameter);
+	} else {
+		cp = cip->buf;
+		cp[cip->bufSize - 1] = '\0';
+		(void) Strncpy(cip->buf, cdCwd, cip->bufSize);
+		if (cp[cip->bufSize - 1] != '\0')
+			return (kErrBadParameter);
+		
+		mkd = (flags & kChdirAndMkdir);
+		pwd = (flags & kChdirAndGetCWD);
+
+		if ((cdCwd[0] == '\0') || (strcmp(cdCwd, ".") == 0)) {
+			result = 0;
+			if (flags == kChdirAndGetCWD)
+				result = FTPGetCWD(cip, newCwd, newCwdSize);
+			return (result);
+		}
+
+		lastSubDir = 0;
+		do {
+			startcp = cp;
+			cp = StrFindLocalPathDelim(cp);
+			if (cp != NULL) {
+				/* If this is the first slash in an absolute
+				 * path, then startcp will be empty.  We will
+				 * use this below to treat this as the root
+				 * directory.
+				 */
+				*cp++ = '\0';
+			} else {
+				lastSubDir = 1;
+			}
+			if (strcmp(startcp, ".") == 0) {
+				result = 0;
+				if ((lastSubDir != 0) && (pwd != 0))
+					result = FTPGetCWD(cip, newCwd, newCwdSize);
+			} else if ((lastSubDir != 0) && (pwd != 0)) {
+				result = FTPChdirAndGetCWD(cip, (*startcp != '\0') ? startcp : "/", newCwd, newCwdSize);
+			} else {
+				result = FTPChdir(cip, (*startcp != '\0') ? startcp : "/");
+			}
+			if (result < 0) {
+				if ((mkd != 0) && (*startcp != '\0')) {
+					if (FTPCmd(cip, "MKD %s", startcp) == 2) {
+						result = FTPChdir(cip, startcp);
+					} else {
+						/* couldn't change nor create */
+						cip->errNo = result;
+					}
+				} else {
+					cip->errNo = result;
+				}
+			}
+		} while ((!lastSubDir) && (result == 0));
+	}
+
+	if (flags == kChdirAndGetCWD) {
+		return (FTPChdirAndGetCWD(cip, cdCwd, newCwd, newCwdSize));
+	} else if (flags == kChdirAndMkdir) {
+		result = FTPMkdir(cip, cdCwd, kRecursiveYes);
+		if (result == kNoErr)
+			result = FTPChdir(cip, cdCwd);
+		return result;
+	} else if (flags == (kChdirAndMkdir|kChdirAndGetCWD)) {
+		result = FTPMkdir(cip, cdCwd, kRecursiveYes);
+		if (result == kNoErr)
+			result = FTPChdirAndGetCWD(cip, cdCwd, newCwd, newCwdSize);
+		return result;
+	} else if ((flags & kChdirOneSubdirAtATime) != 0) {
+		cp = cip->buf;
+		cp[cip->bufSize - 1] = '\0';
+		(void) Strncpy(cip->buf, cdCwd, cip->bufSize);
+		if (cp[cip->bufSize - 1] != '\0')
+			return (kErrBadParameter);
+		
+		mkd = (flags & kChdirAndMkdir);
+		pwd = (flags & kChdirAndGetCWD);
+
+		if ((cdCwd[0] == '\0') || (strcmp(cdCwd, ".") == 0)) {
+			result = 0;
+			if (flags == kChdirAndGetCWD)
+				result = FTPGetCWD(cip, newCwd, newCwdSize);
+			return (result);
+		}
+
+		lastSubDir = 0;
+		do {
+			startcp = cp;
+			cp = StrFindLocalPathDelim(cp);
+			if (cp != NULL) {
+				/* If this is the first slash in an absolute
+				 * path, then startcp will be empty.  We will
+				 * use this below to treat this as the root
+				 * directory.
+				 */
+				*cp++ = '\0';
+			} else {
+				lastSubDir = 1;
+			}
+			if (strcmp(startcp, ".") == 0) {
+				result = 0;
+				if ((lastSubDir != 0) && (pwd != 0))
+					result = FTPGetCWD(cip, newCwd, newCwdSize);
+			} else if ((lastSubDir != 0) && (pwd != 0)) {
+				result = FTPChdirAndGetCWD(cip, (*startcp != '\0') ? startcp : "/", newCwd, newCwdSize);
+			} else {
+				result = FTPChdir(cip, (*startcp != '\0') ? startcp : "/");
+			}
+			if (result < 0) {
+				if ((mkd != 0) && (*startcp != '\0')) {
+					if (FTPCmd(cip, "MKD %s", startcp) == 2) {
+						result = FTPChdir(cip, startcp);
+					} else {
+						/* couldn't change nor create */
+						cip->errNo = result;
+					}
+				} else {
+					cip->errNo = result;
+				}
+			}
+		} while ((!lastSubDir) && (result == 0));
+
+		return (result);
+	}
+
+	return (kErrBadParameter);
+}	/* FTPChdir3 */
+
+
+
+int
+FTPChdirList(FTPCIPtr cip, LineListPtr const cdlist, char *const newCwd, const size_t newCwdSize, int flags)
+{
+	size_t len;
+	char *cdstr;
+	LinePtr lp;
+	int lastSubDir;
+	int mkd, pwd;
+	int result;
+
+	/* Retain backwards compatibility with earlier library versions. */
+	if (flags == kChdirOnly)
+		flags = kChdirFullPath;
+
+	if ((flags & kChdirFullPath) != 0) {
+		len = 0;
+		for (lp = cdlist->first; lp != NULL; lp = lp->next) {
+			len += strlen(lp->line);
+			len++;	/* account for delimiting slash */
+		}
+		cdstr = malloc(len);
+		if (cdstr == NULL)
+			return (cip->errNo = kErrMallocFailed);
+		cdstr[0] = '\0';
+		for (lp = cdlist->first; lp != NULL; lp = lp->next) {
+			strcat(cdstr, lp->line);
+			if (lp->next != NULL)
+				strcat(cdstr, "/");
+		}
+		if (FTPChdir3(cip, cdstr, newCwd, newCwdSize, (flags & ~kChdirOneSubdirAtATime)) == kNoErr) {
+			free(cdstr);
+			return (kNoErr);
+		}
+		free(cdstr);
+	}
+
+	if ((flags & kChdirOneSubdirAtATime) != 0) {
+		mkd = (flags & kChdirAndMkdir);
+		pwd = (flags & kChdirAndGetCWD);
+		lastSubDir = 0;
+		result = kNoErr;
+
+		for (lp = cdlist->first; lp != NULL; lp = lp->next) {
+			if (lp->next == NULL)
+				lastSubDir = 1;
+
+			if (strcmp(lp->line, ".") == 0) {
+				result = 0;
+				if ((lastSubDir != 0) && (pwd != 0))
+					result = FTPGetCWD(cip, newCwd, newCwdSize);
+			} else if ((lastSubDir != 0) && (pwd != 0)) {
+				result = FTPChdirAndGetCWD(cip, (*lp->line != '\0') ? lp->line : "/", newCwd, newCwdSize);
+			} else {
+				result = FTPChdir(cip, (*lp->line != '\0') ? lp->line : "/");
+			}
+			if (result < 0) {
+				if ((mkd != 0) && (*lp->line != '\0')) {
+					if (FTPCmd(cip, "MKD %s", lp->line) == 2) {
+						result = FTPChdir(cip, lp->line);
+					} else {
+						/* couldn't change nor create */
+						cip->errNo = result;
+					}
+				} else {
+					cip->errNo = result;
+				}
+			}
+			if (result != kNoErr)
+				break;
+		}
+		return (result);
+	}
+
+	return (kErrBadParameter);
+}	/* FTPChdirList */
+
+
+
+
 int
 FTPMkdir2(const FTPCIPtr cip, const char *const newDir, const int recurse, const char *const curDir)
 {
@@ -474,6 +766,10 @@ FTPMkdir2(const FTPCIPtr cip, const char *const newDir, const int recurse, const
 					cip->errNo = kErrMKDFailed;
 					return (result);
 				}
+				/* Note: below we will refer to cp + 1
+				 * which is why we set cp to point to
+				 * the byte before the array begins!
+				 */
 				cp = dir - 1;
 				break;
 			}
@@ -570,11 +866,14 @@ FTPFileModificationTime(const FTPCIPtr cip, const char *const file, time_t *cons
 			if (result < 0) {
 				DoneWithResponse(cip, rp);
 				return (result);
-			} else if (strncmp(rp->msg.first->line, "19100", 5) == 0) {
-				Error(cip, kDontPerror, "Warning: Server has Y2K Bug in \"MDTM\" command.\n");
-				cip->errNo = kErrMDTMFailed;
-				result = kErrMDTMFailed;
-			} else if (result == 2) {
+			}
+			if (result == 2) {
+				if (strncmp(rp->msg.first->line, "1910", 4) == 0) {
+					/* Year was printed as "19100" rather than
+					 * "2000" ...
+					 */
+					Error(cip, kDontPerror, "Warning: Server has Y2K Bug in \"MDTM\" command.\n");
+				}
 				*mdtm = UnMDTMDate(rp->msg.first->line);
 				cip->hasMDTM = kCommandAvailable;
 				result = kNoErr;
@@ -1107,6 +1406,11 @@ FTPFileExistsNlst(const FTPCIPtr cip, const char *const file)
 	int result;
 	LineList fileList, rootFileList;
 	char savedCwd[512];
+	int createdTempFile;
+#define kFTPFileExistsNlstTestMessage "This file was created by an FTP client program using the LibNcFTP toolkit.  A temporary file needed to be created to ensure that this directory was not empty, so that the directory could be listed with the guarantee of at least one file in the listing.\r\n\r\nYou may delete this file manually if your FTP client program does not delete it for you.\r\n"
+	const char *testFileMessage = kFTPFileExistsNlstTestMessage; 
+#undef kFTPFileExistsNlstTestMessage
+	const char *testFileName = "testfile.ftp";
 
 	if (cip == NULL)
 		return (kErrBadParameter);
@@ -1171,17 +1475,29 @@ FTPFileExistsNlst(const FTPCIPtr cip, const char *const file)
 		 * anyway.  Then we do the first item
 		 * we found to see if NLST says it exists.
 		 */
+		createdTempFile = 0;
 		if (
 			((result = FTPListToMemory2(cip, "", &rootFileList, "", 0, (int *) 0)) < 0) ||
 			(rootFileList.last == NULL) ||
 			(rootFileList.last->line == NULL)
 		) {
-			/* Hmmm... well, in any case we can't use NLST. */
-			cip->NLSTfileParamWorks = kCommandNotAvailable;
-			cip->errNo = result = kErrNLSTwithFileNotAvailable;
-			DisposeLineListContents(&rootFileList);
-			(void) FTPChdir(cip, savedCwd);
-			return (result);
+			if (AddLine(&rootFileList, testFileName) != NULL) {
+				/* Hate to do this, but if the directory
+				 * is empty but writable (i.e. a $HOME)
+				 * then we can still continue.
+				 */
+				if (FTPPutFileFromMemory(cip, testFileName, testFileMessage, strlen(testFileMessage), kAppendNo) == kNoErr) {
+					createdTempFile = 1;
+				}
+			}
+			if (createdTempFile == 0) {
+				/* Hmmm... well, in any case we can't use NLST. */
+				cip->NLSTfileParamWorks = kCommandNotAvailable;
+				cip->errNo = result = kErrNLSTwithFileNotAvailable;
+				DisposeLineListContents(&rootFileList);
+				(void) FTPChdir(cip, savedCwd);
+				return (result);
+			}
 		}
 
 		if (
@@ -1194,6 +1510,8 @@ FTPFileExistsNlst(const FTPCIPtr cip, const char *const file)
 
 		) {
 			/* Good.  We listed the item. */
+			if (createdTempFile != 0)
+				(void) FTPDelete(cip, testFileName, kRecursiveNo, kGlobNo);
 			DisposeLineListContents(&fileList);
 			DisposeLineListContents(&rootFileList);
 			cip->NLSTfileParamWorks = kCommandAvailable;
@@ -1201,6 +1519,8 @@ FTPFileExistsNlst(const FTPCIPtr cip, const char *const file)
 			/* Don't forget to change back to the original directory. */
 			(void) FTPChdir(cip, savedCwd);
 		} else {
+			if (createdTempFile != 0)
+				(void) FTPDelete(cip, testFileName, kRecursiveNo, kGlobNo);
 			cip->NLSTfileParamWorks = kCommandNotAvailable;
 			cip->errNo = result = kErrNLSTwithFileNotAvailable;
 			DisposeLineListContents(&fileList);
