@@ -1,6 +1,6 @@
 /* cmds.c
  *
- * Copyright (c) 1992-2000 by Mike Gleason.
+ * Copyright (c) 1992-2001 by Mike Gleason.
  * All rights reserved.
  * 
  */
@@ -37,6 +37,13 @@ char gRemoteCWD[512];
  */
 char gPrevRemoteCWD[512];
 
+/* Another buffer we use just temporarily when switching directories. */
+char gScratchCWD[512];
+
+/* If true, the remote server is most likely Windows/DOS.
+ */
+int gServerUsesMSDOSPaths;
+
 /* The only reason we do this is to get gcc/lint to shut up
  * about unused parameters.
  */
@@ -411,14 +418,15 @@ PrintResp(LineListPtr llp)
  * otherwise do a CWD command followed by a PWD.
  */
 int
-nFTPChdirAndGetCWD(const FTPCIPtr cip, const char *cdCwd)
+nFTPChdirAndGetCWD(const FTPCIPtr cip, const char *cdCwd, const int quietMode)
 {
 	ResponsePtr rp;
-	char *l, *r;
 	size_t cdCwdLen;
-	char *cp;
-	char newCdCwd[256];
-	int foundcwd, result;
+	int result;
+#ifdef USE_WHAT_SERVER_SAYS_IS_CWD
+	int foundcwd;
+	char *l, *r;
+#endif
 
 	if (cip == NULL)
 		return (kErrBadParameter);
@@ -438,24 +446,12 @@ nFTPChdirAndGetCWD(const FTPCIPtr cip, const char *cdCwd)
 			cdCwdLen = strlen(cdCwd);
 			if (strcmp(cdCwd, "..") == 0) {
 				result = RCmd(cip, rp, "CDUP"); 	
-			} else if ((strcmp(cdCwd, "-") == 0) && (gPrevRemoteCWD[0] != '\0')) {
-				result = RCmd(cip, rp, "CWD %s", gPrevRemoteCWD);
-			} else if ((cdCwd[cdCwdLen - 1] == '/') && (cdCwdLen > 1)) {
-				/* Strip off trailing slashes that Readline
-				 * may have added.
-				 */
-				STRNCPY(newCdCwd, cdCwd);
-				cp = strrchr(newCdCwd, '/');
-				if ((cp != NULL) && (cp[1] == '\0')) {
-					while ((*cp == '/') && (cp > newCdCwd))
-						*cp-- = '\0';
-				}
-				result = RCmd(cip, rp, "CWD %s", newCdCwd);
 			} else {
 				result = RCmd(cip, rp, "CWD %s", cdCwd);
 			}
 			if (result == 2) {
-				(void) STRNCPY(gPrevRemoteCWD, gRemoteCWD);
+#ifdef USE_WHAT_SERVER_SAYS_IS_CWD
+				(void) STRNCPY(gScratchCWD, gRemoteCWD);
 				foundcwd = 0;
 				if ((r = strrchr(rp->msg.first->line, '"')) != NULL) {
 					/* "xxxx" is current directory.
@@ -473,15 +469,26 @@ nFTPChdirAndGetCWD(const FTPCIPtr cip, const char *cdCwd)
 						result = kNoErr;
 					}
 				}
+				if (quietMode)
+					rp->printMode |= kResponseNoPrint;
 				NcFTPCdResponseProc(cip, rp);
 				DoneWithResponse(cip, rp);
 				if (foundcwd == 0) {
 					result = FTPGetCWD(cip, gRemoteCWD, sizeof(gRemoteCWD));
 					if (result != kNoErr) {
-						PathCat(gRemoteCWD, sizeof(gRemoteCWD), gPrevRemoteCWD, cdCwd);
+						PathCat(gRemoteCWD, sizeof(gRemoteCWD), gScratchCWD, cdCwd, gServerUsesMSDOSPaths);
 						result = kNoErr;
 					}
 				}
+#else /* USE_CLIENT_SIDE_CALCULATED_CWD */
+				if (quietMode)
+					rp->printMode |= kResponseNoPrint;
+				NcFTPCdResponseProc(cip, rp);
+				DoneWithResponse(cip, rp);
+				(void) STRNCPY(gScratchCWD, gRemoteCWD);
+				PathCat(gRemoteCWD, sizeof(gRemoteCWD), gScratchCWD, cdCwd, gServerUsesMSDOSPaths);
+				result = kNoErr;
+#endif
 			} else if (result > 0) {
 				result = kErrCWDFailed;
 				cip->errNo = kErrCWDFailed;
@@ -497,6 +504,64 @@ nFTPChdirAndGetCWD(const FTPCIPtr cip, const char *cdCwd)
 
 
 
+int
+Chdirs(FTPCIPtr cip, const char *const cdCwd)
+{
+	char *cp, *startcp;
+	int result;
+	int lastSubDir;
+
+	if (cip == NULL)
+		return (kErrBadParameter);
+	if (strcmp(cip->magic, kLibraryMagic))
+		return (kErrBadMagic);
+
+	if (cdCwd == NULL) {
+		result = kErrInvalidDirParam;
+		cip->errNo = kErrInvalidDirParam;
+		return result;
+	}
+	
+	if ((cdCwd[0] == '\0') || (strcmp(cdCwd, ".") == 0))
+		return (kNoErr);
+
+	cp = cip->buf;
+	cp[cip->bufSize - 2] = '\0';
+	if ((cdCwd[0] == '.') && (cdCwd[1] == '.') && ((cdCwd[2] == '\0') || IsLocalPathDelim(cdCwd[2]))) {
+		PathCat(cip->buf, cip->bufSize, gRemoteCWD, cdCwd, gServerUsesMSDOSPaths);
+	} else {
+		(void) Strncpy(cip->buf, cdCwd, cip->bufSize);
+	}
+	if (cp[cip->bufSize - 2] != '\0')
+		return (kErrBadParameter);
+
+	StrRemoveTrailingLocalPathDelim(cp);
+
+	/* Try doing the whole path at once. */
+	result = nFTPChdirAndGetCWD(cip, cp, 1);
+	if (result == kNoErr)
+		return (result);
+
+	/* Try doing the path one subdir at a time. */
+	do {
+		startcp = cp;
+		cp = StrFindLocalPathDelim(cp + 0);
+		if (cp != NULL) {
+			*cp++ = '\0';
+		}
+		lastSubDir = (cp == NULL);
+		result = nFTPChdirAndGetCWD(cip, (*startcp != '\0') ? startcp : "/", lastSubDir ? 0 : 1);
+		if (result < 0) {
+			cip->errNo = result;
+		}
+	} while ((!lastSubDir) && (result == kNoErr));
+
+	return (result);
+}	/* Chdirs */
+
+
+
+
 /* Remote change of working directory command. */
 void
 ChdirCmd(const int argc, const char **const argv, const CommandPtr cmdp, const ArgvInfoPtr aip)
@@ -506,19 +571,46 @@ ChdirCmd(const int argc, const char **const argv, const CommandPtr cmdp, const A
 	LinePtr lp;
 
 	ARGSUSED(gUnusedArg);
-	InitLineList(&ll);
-	result = FTPRemoteGlob(&gConn, &ll, argv[1], (aip->noglobargv[1] != 0) ? kGlobNo: kGlobYes);
-	if (result < 0) {
-		FTPPerror(&gConn, result, kErrGlobFailed, argv[0], argv[1]);
-	} else {
-		lp = ll.first;
-		if ((lp != NULL) && (lp->line != NULL)) {
-			result = nFTPChdirAndGetCWD(&gConn, lp->line);
-			if (result != kNoErr)
-				FTPPerror(&gConn, result, kErrCWDFailed, "Could not chdir to", lp->line);
+
+	if (argc <= 1) {
+		if (gStartDir[0] != '\0') {
+			(void) STRNCPY(gPrevRemoteCWD, gRemoteCWD);
+			result = Chdirs(&gConn, gStartDir);
+			if (result != kNoErr) {
+				/* State is incoherent if this happens! */
+				FTPPerror(&gConn, result, kErrCWDFailed, "Could not chdir to", gStartDir);
+			}
+		} else {
+			PrintCmdUsage(cmdp);
 		}
+	} else {
+		InitLineList(&ll);
+		result = FTPRemoteGlob(&gConn, &ll, argv[1], (aip->noglobargv[1] != 0) ? kGlobNo: kGlobYes);
+		if (result < 0) {
+			FTPPerror(&gConn, result, kErrGlobFailed, argv[0], argv[1]);
+		} else {
+			lp = ll.first;
+			if ((lp != NULL) && (lp->line != NULL)) {
+				if ((strcmp(lp->line, "-") == 0) && (gPrevRemoteCWD[0] != '\0')) {
+					free(lp->line);
+					lp->line = StrDup(gPrevRemoteCWD);
+					if (lp->line == NULL) {
+						result = kErrMallocFailed;
+						gConn.errNo = kErrMallocFailed;
+					} else {
+						(void) STRNCPY(gPrevRemoteCWD, gRemoteCWD);
+						result = Chdirs(&gConn, lp->line);
+					}
+				} else {
+					(void) STRNCPY(gPrevRemoteCWD, gRemoteCWD);
+					result = Chdirs(&gConn, lp->line);
+				}
+				if (result != kNoErr)
+					FTPPerror(&gConn, result, kErrCWDFailed, "Could not chdir to", lp->line);
+			}
+		}
+		DisposeLineListContents(&ll);
 	}
-	DisposeLineListContents(&ll);
 }	/* ChdirCmd */
 
 
@@ -674,7 +766,7 @@ NcFTPConfirmResumeDownloadProc(
 
 	tstr[sizeof(tstr) - 1] = '\0';
 	(void) strftime(tstr, sizeof(tstr) - 1, "%c", localtime((time_t *) &localmtime)); 
-	(void) printf(
+	(void) fprintf(stdout,
 #if defined(HAVE_LONG_LONG) && defined(PRINTF_LONG_LONG_LLD)
 		"\nThe local file \"%s\" already exists.\n\tLocal:  %12lld bytes, dated %s.\n",
 #elif defined(HAVE_LONG_LONG) && defined(PRINTF_LONG_LONG_QD)
@@ -691,7 +783,7 @@ NcFTPConfirmResumeDownloadProc(
 
 	if ((remotemtime != kModTimeUnknown) && (remotesize != kSizeUnknown)) {
 		(void) strftime(tstr, sizeof(tstr) - 1, "%c", localtime((time_t *) &remotemtime)); 
-		(void) printf(
+		(void) fprintf(stdout,
 #if defined(HAVE_LONG_LONG) && defined(PRINTF_LONG_LONG_LLD)
 			"\tRemote: %12lld bytes, dated %s.\n",
 #elif defined(HAVE_LONG_LONG) && defined(PRINTF_LONG_LONG_QD)
@@ -709,7 +801,7 @@ NcFTPConfirmResumeDownloadProc(
 			return (kConfirmResumeProcSaidSkip);
 		}
 	} else if (remotesize != kSizeUnknown) {
-		(void) printf(
+		(void) fprintf(stdout,
 #if defined(HAVE_LONG_LONG) && defined(PRINTF_LONG_LONG_LLD)
 			"\tRemote: %12lld bytes, date unknown.\n",
 #elif defined(HAVE_LONG_LONG) && defined(PRINTF_LONG_LONG_QD)
@@ -855,6 +947,8 @@ GetCmd(const int argc, const char **const argv, const CommandPtr cmdp, const Arg
 	GetoptReset();
 	while ((opt = Getopt(argc, argv, "aAzfrRTD")) >= 0) switch (opt) {
 		case 'a':
+			xtype = kTypeAscii;
+			break;
 		case 'A':
 			/* Append to local files, instead of truncating
 			 * them first.
@@ -1373,8 +1467,10 @@ ListCmd(const int argc, const char **const argv, const CommandPtr cmdp, const Ar
 
 #if defined(WIN32) || defined(_WINDOWS)
 #elif defined(HAVE_SIGSETJMP)
+		osigpipe = osigint = (sigproc_t) 0;
 		sj = sigsetjmp(gCancelJmp, 1);
 #else	/* HAVE_SIGSETJMP */
+		osigpipe = osigint = (sigproc_t) 0;
 		sj = setjmp(gCancelJmp);
 #endif	/* HAVE_SIGSETJMP */
 
@@ -1384,15 +1480,15 @@ ListCmd(const int argc, const char **const argv, const CommandPtr cmdp, const Ar
 			/* Caught a signal. */
 			(void) NcSignal(SIGPIPE, (FTPSigProc) SIG_IGN);
 			ClosePager(stream);
-			(void) NcSignal(SIGPIPE, (sigproc_t) osigpipe);
-			(void) NcSignal(SIGINT, (sigproc_t) osigint);
+			(void) NcSignal(SIGPIPE, osigpipe);
+			(void) NcSignal(SIGINT, osigint);
 			Trace(0, "Canceled because of signal %d.\n", gGotSig);
 			(void) fprintf(stderr, "Canceled.\n");
 			gMayCancelJmp = 0;
 			return;
 		} else {
-			osigpipe = (vsigproc_t) NcSignal(SIGPIPE, Cancel);
-			osigint = (vsigproc_t) NcSignal(SIGINT, Cancel);
+			osigpipe = NcSignal(SIGPIPE, Cancel);
+			osigint = NcSignal(SIGINT, Cancel);
 			gMayCancelJmp = 1;
 		}
 #endif
@@ -1412,8 +1508,8 @@ ListCmd(const int argc, const char **const argv, const CommandPtr cmdp, const Ar
 		ClosePager(stream);
 #if defined(WIN32) || defined(_WINDOWS)
 #else
-		(void) NcSignal(SIGPIPE, (sigproc_t) osigpipe);
-		(void) NcSignal(SIGINT, (sigproc_t) osigint);
+		(void) NcSignal(SIGPIPE, osigpipe);
+		(void) NcSignal(SIGINT, osigint);
 #endif
 	}
 	gMayCancelJmp = 0;
@@ -1586,15 +1682,15 @@ LocalListCmd(const int argc, const char **const argv, const CommandPtr cmdp, con
 		ClosePager(outfp);
 		if (infp != NULL)
 			(void) pclose(infp);
-		(void) NcSignal(SIGPIPE, (sigproc_t) osigpipe);
-		(void) NcSignal(SIGINT, (sigproc_t) osigint);
+		(void) NcSignal(SIGPIPE, osigpipe);
+		(void) NcSignal(SIGINT, osigint);
 		(void) fprintf(stderr, "Canceled.\n");
 		Trace(0, "Canceled because of signal %d.\n", gGotSig);
 		gMayCancelJmp = 0;
 		return;
 	} else {
-		osigpipe = (vsigproc_t) NcSignal(SIGPIPE, Cancel);
-		osigint = (vsigproc_t) NcSignal(SIGINT, Cancel);
+		osigpipe = NcSignal(SIGPIPE, Cancel);
+		osigint = NcSignal(SIGINT, Cancel);
 		gMayCancelJmp = 1;
 	}
 
@@ -1607,8 +1703,8 @@ LocalListCmd(const int argc, const char **const argv, const CommandPtr cmdp, con
 	ClosePager(outfp);
 	outfp = NULL;
 
-	(void) NcSignal(SIGPIPE, (sigproc_t) osigpipe);
-	(void) NcSignal(SIGINT, (sigproc_t) osigint);
+	(void) NcSignal(SIGPIPE, osigpipe);
+	(void) NcSignal(SIGINT, osigint);
 	gMayCancelJmp = 0;
 #endif
 }	/* LocalListCmd */
@@ -2060,10 +2156,6 @@ static void
 NcFTPGetPassphraseProc(const FTPCIPtr cip, LineListPtr pwPrompt, char *pass, size_t dsize)
 {
 	LinePtr lp;
-#if defined(WIN32) || defined(_WINDOWS)
-#else
-	char *password;
-#endif
 
 	(void) printf("\nPassword requested by %s for user \"%s\".\n\n",
 		cip->host,
@@ -2074,24 +2166,7 @@ NcFTPGetPassphraseProc(const FTPCIPtr cip, LineListPtr pwPrompt, char *pass, siz
 		(void) printf("    %s\n", lp->line);
 	}
 	(void) printf("\n");
-
-#if defined(WIN32) || defined(_WINDOWS)
-	(void) gl_win_getpass("Password: ", pass, dsize);
-#elif defined(HAVE_GETPASSPHRASE)
-	password = getpassphrase("Password: ");
-	if (password != NULL) {
-		(void) Strncpy(pass, password, dsize);
-		/* Avoid cleartext password in memory. */
-		(void) memset(password, 0, strlen(cip->pass));
-	}
-#else	/* HAVE_GETPASSPHRASE */
-	password = GetPass("Password: ");
-	if (password != NULL) {
-		(void) Strncpy(pass, password, dsize);
-		/* Avoid cleartext password in memory. */
-		(void) memset(password, 0, strlen(cip->pass));
-	}
-#endif	/* HAVE_GETPASSPHRASE */
+	(void) gl_getpass("Password: ", pass, (int) dsize);
 }	/* NcFTPGetPassphraseProc */
 
 
@@ -2107,8 +2182,8 @@ DoOpen(void)
 #ifdef SIGALRM
 	sigproc_t osigalrm;
 #endif
-	char *password;
 	char prompt[256];
+	char *cp;
 
 	if (gConn.firewallType == kFirewallNotInUse) {
 		(void) STRNCPY(ohost, gConn.host);
@@ -2167,24 +2242,7 @@ DoOpen(void)
 				(void) STRNCAT(prompt, "\" at ");
 				(void) STRNCAT(prompt, gConn.firewallHost);
 				(void) STRNCAT(prompt, ": ");
-
-#if defined(WIN32) || defined(_WINDOWS)
-				(void) gl_win_getpass(prompt, gConn.firewallPass, sizeof(gConn.firewallPass));
-#elif defined(HAVE_GETPASSPHRASE)
-				password = getpassphrase(prompt);
-				if (password != NULL) {
-					(void) STRNCPY(gConn.firewallPass, password);
-					/* Avoid cleartext password in memory. */
-					(void) memset(password, 0, strlen(gConn.firewallPass));
-				}
-#else	/* HAVE_GETPASSPHRASE */
-				password = GetPass(prompt);
-				if (password != NULL) {
-					(void) STRNCPY(gConn.firewallPass, password);
-					/* Avoid cleartext password in memory. */
-					(void) memset(password, 0, strlen(gConn.firewallPass));
-				}
-#endif	/* HAVE_GETPASSPHRASE */
+				(void) gl_getpass(prompt, gConn.firewallPass, sizeof(gConn.firewallPass));
 				break;
 		}
 	}
@@ -2219,16 +2277,20 @@ DoOpen(void)
 		gConn.printResponseProc = 0;
 
 		/* Need to note what our "root" was before we change it. */
+		gServerUsesMSDOSPaths = 0;
 		if (gConn.startingWorkingDirectory == NULL) {
 			(void) STRNCPY(gRemoteCWD, "/");	/* Guess! */
 		} else {
-			(void) STRNCPY(gRemoteCWD, gConn.startingWorkingDirectory);
+			cp = gConn.startingWorkingDirectory;
+			(void) STRNCPY(gRemoteCWD, cp);
+			if ((strlen(cp) >= 2) && (isalpha(cp[0])) && (cp[1] == ':'))
+				gServerUsesMSDOSPaths = 1;
 		}
 		(void) STRNCPY(gStartDir, gRemoteCWD);
 
 		/* If the bookmark specified a remote directory, change to it now. */
 		if ((gLoadedBm != 0) && (gBm.dir[0] != '\0')) {
-			result = nFTPChdirAndGetCWD(&gConn, gBm.dir);
+			result = Chdirs(&gConn, gBm.dir);
 			if (result < 0) {
 				FTPPerror(&gConn, result, kErrCWDFailed, "Could not chdir to previous directory", gBm.dir);
 			}
@@ -2278,7 +2340,6 @@ OpenCmd(const int argc, const char **const argv, const CommandPtr cmdp, const Ar
 	int directoryURL = 0;
 	LineList cdlist;
 	LinePtr lp;
-	char *password;
 	char prompt[256];
 
 	ARGSUSED(gUnusedArg);
@@ -2377,8 +2438,9 @@ OpenCmd(const int argc, const char **const argv, const CommandPtr cmdp, const Ar
 	}
 
 	GetoptReset();
-	while ((c = Getopt(argc, argv, "aP:u:p:J:rd:g:")) > 0) switch(c) {
+	while ((c = Getopt(argc, argv, "aP:u:p:J:j:rd:g:")) > 0) switch(c) {
 		case 'J':
+		case 'j':
 			(void) STRNCPY(gConn.acct, gOptArg);
 			break;
 		case 'a':
@@ -2418,24 +2480,7 @@ OpenCmd(const int argc, const char **const argv, const CommandPtr cmdp, const Ar
 		(void) STRNCPY(prompt, "Username at ");
 		(void) STRNCAT(prompt, gConn.host);
 		(void) STRNCAT(prompt, ": ");
-
-#if defined(WIN32) || defined(_WINDOWS)
-		(void) gl_win_getpass(prompt, gConn.user, sizeof(gConn.user));
-#elif defined(HAVE_GETPASSPHRASE)
-		password = getpassphrase(prompt);
-		if (password != NULL) {
-			(void) STRNCPY(gConn.user, password);
-			/* Avoid cleartext password in memory. */
-			(void) memset(password, 0, strlen(gConn.user));
-		}
-#else	/* HAVE_GETPASSPHRASE */
-		password = GetPass(prompt);
-		if (password != NULL) {
-			(void) STRNCPY(gConn.user, password);
-			/* Avoid cleartext password in memory. */
-			(void) memset(password, 0, strlen(gConn.user));
-		}
-#endif	/* HAVE_GETPASSPHRASE */
+		(void) gl_getpass(prompt, gConn.user, sizeof(gConn.user));
 	}
 
 	rc = DoOpen();
@@ -2483,8 +2528,10 @@ PageCmd(const int argc, const char **const argv, const CommandPtr cmdp, const Ar
 #else
 
 #ifdef HAVE_SIGSETJMP
+	osigpipe = osigint = (sigproc_t) 0;
 	sj = sigsetjmp(gCancelJmp, 1);
 #else	/* HAVE_SIGSETJMP */
+	osigpipe = osigint = (sigproc_t) 0;
 	sj = setjmp(gCancelJmp);
 #endif	/* HAVE_SIGSETJMP */
 
@@ -2492,15 +2539,15 @@ PageCmd(const int argc, const char **const argv, const CommandPtr cmdp, const Ar
 		/* Caught a signal. */
 		(void) NcSignal(SIGPIPE, (FTPSigProc) SIG_IGN);
 		ClosePager(stream);
-		(void) NcSignal(SIGPIPE, (sigproc_t) osigpipe);
-		(void) NcSignal(SIGINT, (sigproc_t) osigint);
+		(void) NcSignal(SIGPIPE, osigpipe);
+		(void) NcSignal(SIGINT, osigint);
 		(void) fprintf(stderr, "Canceled.\n");
 		Trace(0, "Canceled because of signal %d.\n", gGotSig);
 		gMayCancelJmp = 0;
 		return;
 	} else {
-		osigpipe = (vsigproc_t) NcSignal(SIGPIPE, Cancel);
-		osigint = (vsigproc_t) NcSignal(SIGINT, Cancel);
+		osigpipe = NcSignal(SIGPIPE, Cancel);
+		osigint = NcSignal(SIGINT, Cancel);
 		gMayCancelJmp = 1;
 	}
 
@@ -2523,8 +2570,8 @@ PageCmd(const int argc, const char **const argv, const CommandPtr cmdp, const Ar
 #else
 	(void) NcSignal(SIGPIPE, (FTPSigProc) SIG_IGN);
 	ClosePager(stream);
-	(void) NcSignal(SIGPIPE, (sigproc_t) osigpipe);
-	(void) NcSignal(SIGINT, (sigproc_t) osigint);
+	(void) NcSignal(SIGPIPE, osigpipe);
+	(void) NcSignal(SIGINT, osigint);
 #endif
 	gMayCancelJmp = 0;
 }	/* PageCmd */
@@ -2556,7 +2603,7 @@ NcFTPConfirmResumeUploadProc(
 
 	if ((localmtime != kModTimeUnknown) && (localsize != kSizeUnknown)) {
 		(void) strftime(tstr, sizeof(tstr) - 1, "%c", localtime((time_t *) &localmtime)); 
-		(void) printf(
+		(void) fprintf(stdout,
 #if defined(HAVE_LONG_LONG) && defined(PRINTF_LONG_LONG_LLD)
 			"\tLocal:  %12lld bytes, dated %s.\n",
 #elif defined(HAVE_LONG_LONG) && defined(PRINTF_LONG_LONG_QD)
@@ -2574,7 +2621,7 @@ NcFTPConfirmResumeUploadProc(
 			return (kConfirmResumeProcSaidSkip);
 		}
 	} else if (localsize != kSizeUnknown) {
-		(void) printf(
+		(void) fprintf(stdout,
 #if defined(HAVE_LONG_LONG) && defined(PRINTF_LONG_LONG_LLD)
 			"\tLocal:  %12lld bytes, date unknown.\n",
 #elif defined(HAVE_LONG_LONG) && defined(PRINTF_LONG_LONG_QD)
@@ -2597,7 +2644,7 @@ NcFTPConfirmResumeUploadProc(
 	tstr[sizeof(tstr) - 1] = '\0';
 	if ((remotemtime != kModTimeUnknown) && (remotesize != kSizeUnknown)) {
 		(void) strftime(tstr, sizeof(tstr) - 1, "%c", localtime((time_t *) &remotemtime)); 
-		(void) printf(
+		(void) fprintf(stdout,
 #if defined(HAVE_LONG_LONG) && defined(PRINTF_LONG_LONG_LLD)
 			"\tRemote: %12lld bytes, dated %s.\n",
 #elif defined(HAVE_LONG_LONG) && defined(PRINTF_LONG_LONG_QD)
@@ -2611,7 +2658,7 @@ NcFTPConfirmResumeUploadProc(
 			tstr
 		);
 	} else if (remotesize != kSizeUnknown) {
-		(void) printf(
+		(void) fprintf(stdout,
 #if defined(HAVE_LONG_LONG) && defined(PRINTF_LONG_LONG_LLD)
 			"\tRemote: %12lld bytes, date unknown.\n",
 #elif defined(HAVE_LONG_LONG) && defined(PRINTF_LONG_LONG_QD)
@@ -2755,8 +2802,10 @@ PutCmd(const int argc, const char **const argv, const CommandPtr cmdp, const Arg
 	ARGSUSED(gUnusedArg);
 	GetoptReset();
 	while ((opt = Getopt(argc, argv, "AafZzrRD")) >= 0) switch (opt) {
-		case 'A':
 		case 'a':
+			xtype = kTypeAscii;
+			break;
+		case 'A':
 			/* Append to remote files, instead of truncating
 			 * them first.
 			 */
@@ -2801,6 +2850,9 @@ PutCmd(const int argc, const char **const argv, const CommandPtr cmdp, const Arg
 			return;
 	}
 
+	if (nD >= 2)
+		deleteFlag = kDeleteYes;
+
 	if (renameMode != 0) {
 		if (gOptInd > (argc - 2)) {
 			PrintCmdUsage(cmdp);
@@ -2839,8 +2891,10 @@ PwdCmd(const int argc, const char **const argv, const CommandPtr cmdp, const Arg
 {
 	int result;
 	char url[256];
+	char olddir[256];
 
 	ARGSUSED(gUnusedArg);
+#ifdef USE_WHAT_SERVER_SAYS_IS_CWD
 	result = FTPGetCWD(&gConn, gRemoteCWD, sizeof(gRemoteCWD));
 	CurrentURL(url, sizeof(url), 0);
 	if (result < 0) {
@@ -2848,6 +2902,31 @@ PwdCmd(const int argc, const char **const argv, const CommandPtr cmdp, const Arg
 	} else {
 		Trace(-1, "%s\n", url);
 	}
+#else /* USE_CLIENT_SIDE_CALCULATED_CWD */
+
+	/* Display the current working directory, as
+	 * maintained by us.
+	 */
+	CurrentURL(url, sizeof(url), 0);
+	Trace(-1, "  %s\n", url);
+	olddir[sizeof(olddir) - 2] = '\0';
+	STRNCPY(olddir, gRemoteCWD);
+
+	/* Now see what the server reports as the CWD.
+	 * Because of symlinks, it could be different
+	 * from what we are using.
+	 */
+	result = FTPGetCWD(&gConn, gRemoteCWD, sizeof(gRemoteCWD));
+	if ((result == kNoErr) && (strcmp(gRemoteCWD, olddir) != 0)) {
+		Trace(-1, "This URL is also valid on this server:\n");
+		CurrentURL(url, sizeof(url), 0);
+		Trace(-1, "  %s\n", url);
+		if (olddir[sizeof(olddir) - 2] == '\0') {
+			/* Go back to using the non-resolved version. */
+			STRNCPY(gRemoteCWD, olddir);
+		}
+	}
+#endif
 }	/* PwdCmd */
 
 
@@ -3207,7 +3286,7 @@ BGStartCmd(const int argc, const char **const argv, const CommandPtr cmdp, const
 		return;
 
 	if ((argc < 2) || ((n = atoi(argv[1])) < 2)) {
-		RunBatch(0, &gConn);
+		RunBatch();
 		(void) printf("Background process started.\n");
 #if defined(WIN32) || defined(_WINDOWS)
 #else
@@ -3215,7 +3294,7 @@ BGStartCmd(const int argc, const char **const argv, const CommandPtr cmdp, const
 #endif
 	} else {
 		for (i=0; i<n; i++)
-			RunBatch(0, &gConn);
+			RunBatch();
 		(void) printf("Background processes started.\n");
 #if defined(WIN32) || defined(_WINDOWS)
 #else
@@ -3226,15 +3305,12 @@ BGStartCmd(const int argc, const char **const argv, const CommandPtr cmdp, const
 
 
 
-
-/* This commands lets the user change the umask, if the server supports it. */
 /* (bgget) */
 void
 SpoolGetCmd(const int argc, const char **const argv, const CommandPtr cmdp, const ArgvInfoPtr aip)
 {
 	int opt;
 	int renameMode = 0;
-	int recurseFlag = kRecursiveNo;
 	int rc;
 	int i;
 	int xtype = gBm.xferType;
@@ -3257,7 +3333,7 @@ SpoolGetCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 	}
 
 	GetoptReset();
-	while ((opt = Getopt(argc, argv, "@:aAzfrRD")) >= 0) switch (opt) {
+	while ((opt = Getopt(argc, argv, "@:azfrD")) >= 0) switch (opt) {
 		case '@':
 			when = GetStartSpoolDate(gOptArg);
 			if ((when == (time_t) -1) || (when == (time_t) 0)) {
@@ -3265,6 +3341,9 @@ SpoolGetCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 				return;
 			}
 			break;
+		case 'a':
+			xtype = kTypeAscii;
+			break;
 		case 'z':
 			/* Special flag that lets you specify the
 			 * destination file.  Normally a "get" will
@@ -3273,13 +3352,6 @@ SpoolGetCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 			 */
 			renameMode = 1;
 			break;
-		case 'r':
-		case 'R':
-			/* If the item is a directory, get the
-			 * directory and all its contents.
-			 */
-			recurseFlag = kRecursiveYes;
-			break;
 		case 'D':
 			/* You can delete the remote file after
 			 * you downloaded it successfully by using
@@ -3308,6 +3380,8 @@ SpoolGetCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 			return;
 		}
 		rc = SpoolX(
+			NULL,
+			NULL,
 			"get",
 			argv[gOptInd],
 			gRemoteCWD,
@@ -3318,14 +3392,21 @@ SpoolGetCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 			gConn.port,
 			gConn.user,
 			gConn.pass,
+			gConn.acct,
 			xtype,
-			recurseFlag,
+			kRecursiveNo,
 			deleteFlag,
 			gConn.dataPortMode,
-			when
+			NULL,
+			NULL,
+			NULL,
+			NULL,
+			NULL,
+			when,
+			0
 		);
 		if (rc == 0) {
-			Trace(-1, "  + Spooled: get %s as %s\n", argv[gOptInd], argv[gOptInd]);
+			Trace(-1, "  + Spooled: get %s as %s\n", argv[gOptInd], argv[gOptInd + 1]);
 		}
 	} else {
 		for (i=gOptInd; i<argc; i++) {
@@ -3336,33 +3417,40 @@ SpoolGetCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 			if (rc < 0) {
 				FTPPerror(&gConn, rc, kErrGlobFailed, argv[0], pattern);
 			} else {
-				for (lp = ll.first; lp != NULL; lp = lp->next) {
-					if (lp->line != NULL) {
-						lname = strrchr(lp->line, '/');
-						if (lname == NULL)
-							lname = lp->line;
-						else
-							lname++;
-						rc = SpoolX(
-							"get",
-							lp->line,
-							gRemoteCWD,
-							lname,
-							ldir,
-							gConn.host,
-							gConn.ip,
-							gConn.port,
-							gConn.user,
-							gConn.pass,
-							xtype,
-							recurseFlag,
-							deleteFlag,
-							gConn.dataPortMode,
-							when
-						);
-						if (rc == 0) {
-							Trace(-1, "  + Spooled: get %s\n", lp->line);
-						}
+				for (lp = ll.first; (lp != NULL) && (lp->line != NULL); lp = lp->next) {
+					lname = strrchr(lp->line, '/');
+					if (lname == NULL)
+						lname = lp->line;
+					else
+						lname++;
+					rc = SpoolX(
+						NULL,
+						NULL,
+						"get",
+						lp->line,
+						gRemoteCWD,
+						lname,
+						ldir,
+						gConn.host,
+						gConn.ip,
+						gConn.port,
+						gConn.user,
+						gConn.pass,
+						gConn.acct,
+						xtype,
+						kRecursiveNo,
+						deleteFlag,
+						gConn.dataPortMode,
+						NULL,
+						NULL,
+						NULL,
+						NULL,
+						NULL,
+						when,
+						0
+					);
+					if (rc == 0) {
+						Trace(-1, "  + Spooled: get %s\n", lp->line);
 					}
 				}
 			}
@@ -3374,14 +3462,12 @@ SpoolGetCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 
 
 
-/* This commands lets the user change the umask, if the server supports it. */
 /* (bgput) */
 void
 SpoolPutCmd(const int argc, const char **const argv, const CommandPtr cmdp, const ArgvInfoPtr aip)
 {
 	int opt;
 	int renameMode = 0;
-	int recurseFlag = kRecursiveNo;
 	int rc;
 	int i;
 	int xtype = gBm.xferType;
@@ -3404,7 +3490,7 @@ SpoolPutCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 	}
 
 	GetoptReset();
-	while ((opt = Getopt(argc, argv, "@:aAzfrRD")) >= 0) switch (opt) {
+	while ((opt = Getopt(argc, argv, "@:azrD")) >= 0) switch (opt) {
 		case '@':
 			when = GetStartSpoolDate(gOptArg);
 			if ((when == (time_t) -1) || (when == (time_t) 0)) {
@@ -3412,6 +3498,9 @@ SpoolPutCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 				return;
 			}
 			break;
+		case 'a':
+			xtype = kTypeAscii;
+			break;
 		case 'z':
 			/* Special flag that lets you specify the
 			 * destination file.  Normally a "get" will
@@ -3420,13 +3509,6 @@ SpoolPutCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 			 */
 			renameMode = 1;
 			break;
-		case 'r':
-		case 'R':
-			/* If the item is a directory, get the
-			 * directory and all its contents.
-			 */
-			recurseFlag = kRecursiveYes;
-			break;
 		case 'D':
 			/* You can delete the remote file after
 			 * you downloaded it successfully by using
@@ -3455,6 +3537,8 @@ SpoolPutCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 			return;
 		}
 		rc = SpoolX(
+			NULL,
+			NULL,
 			"put",
 			argv[gOptInd + 1],
 			gRemoteCWD,
@@ -3465,14 +3549,21 @@ SpoolPutCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 			gConn.port,
 			gConn.user,
 			gConn.pass,
+			gConn.acct,
 			xtype,
-			recurseFlag,
+			kRecursiveNo,
 			deleteFlag,
 			gConn.dataPortMode,
-			when
+			NULL,
+			NULL,
+			NULL,
+			NULL,
+			NULL,
+			when,
+			0
 		);
 		if (rc == 0) {
-			Trace(-1, "  + Spooled: put %s as %s\n", argv[gOptInd], argv[gOptInd]);
+			Trace(-1, "  + Spooled: put %s as %s\n", argv[gOptInd], argv[gOptInd + 1]);
 		}
 	} else {
 		for (i=gOptInd; i<argc; i++) {
@@ -3491,6 +3582,8 @@ SpoolPutCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 						else
 							rname++;
 						rc = SpoolX(
+							NULL,
+							NULL,
 							"put",
 							rname,
 							gRemoteCWD,
@@ -3501,11 +3594,18 @@ SpoolPutCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 							gConn.port,
 							gConn.user,
 							gConn.pass,
+							gConn.acct,
 							xtype,
-							recurseFlag,
+							kRecursiveNo,
 							deleteFlag,
 							gConn.dataPortMode,
-							when
+							NULL,
+							NULL,
+							NULL,
+							NULL,
+							NULL,
+							when,
+							0
 						);
 						if (rc == 0) {
 							Trace(-1, "  + Spooled: put %s\n", lp->line);
@@ -3521,7 +3621,7 @@ SpoolPutCmd(const int argc, const char **const argv, const CommandPtr cmdp, cons
 
 
 
-/* This commands lets the user change the umask, if the server supports it. */
+/* This commands lets the user create symbolic links, if the server supports it. */
 void
 SymlinkCmd(const int argc, const char **const argv, const CommandPtr cmdp, const ArgvInfoPtr aip)
 {
