@@ -1,6 +1,6 @@
 /******************************************************************************/
 
-char gVersion[] = "@(#) NcFTP 3.0.0/427 March 20 2000, 02:05 AM               ";
+char gVersion[] = "@(#) NcFTP 3.0.4/692 October 25 2001, 10:00 PM             ";
 
 /******************************************************************************/
 
@@ -9,13 +9,13 @@ char gOS[] = OS;
 #elif defined(WIN32) || defined(_WINDOWS)
 char gOS[] = "Windows";
 #else
-char gOS[] = "";
+char gOS[] = "UNIX";
 #endif
 
 /******************************************************************************/
 
 char gCopyright[] = "@(#) \
-Copyright (c) 1992-1999 by Mike Gleason.\n\
+Copyright (c) 1992-2001 by Mike Gleason.\n\
 All rights reserved.\n\
 ";
 
